@@ -1,3 +1,7 @@
+"""
+Jobs
+"""
+
 from __future__ import annotations
 
 import dataclasses
@@ -6,27 +10,33 @@ import typing as t
 
 from saq.utils import exponential_backoff, now, seconds, uuid1
 
-ABORT_ID_PREFIX = "saq:abort:"
 
 if t.TYPE_CHECKING:
     from saq.queue import Queue
     from saq.types import DurationKind, Function
 
+ABORT_ID_PREFIX = "saq:abort:"
+
 
 def get_default_job_key() -> str:
     return uuid1()
 
 
 class Status(str, enum.Enum):
+    """
+    Queue Status
+    """
+
     NEW = "new"
-    DEFERRED = "deferred"
     QUEUED = "queued"
     ACTIVE = "active"
+    ABORTING = "aborting"
     ABORTED = "aborted"
     FAILED = "failed"
     COMPLETE = "complete"
 
 
+ACTIVE_STATUSES = {Status.NEW, Status.QUEUED, Status.ACTIVE}
 TERMINAL_STATUSES = {Status.COMPLETE, Status.FAILED, Status.ABORTED}
 UNSUCCESSFUL_TERMINAL_STATUSES = TERMINAL_STATUSES - {Status.COMPLETE}
 
@@ -36,11 +46,17 @@ class CronJob:
     """
     Allows scheduling of repeated jobs with cron syntax.
 
-    function: the async function to run
-    cron: cron string for a job to be repeated, uses croniter
-    unique: unique jobs only one once per queue, defaults true
+    Args:
+        function (saq.types.Function): the async function to run
+        cron (str): cron string for a job to be repeated, uses croniter
+        unique (bool): unique jobs only one once per queue, defaults true
 
-    Remaining kwargs are pass through to Job
+    **The remaining kwargs are pass through to Job:** (see equivalent field in Job for more)
+
+    * timeout
+    * heartbeat
+    * retries
+    * ttl
     """
 
     function: Function
@@ -50,6 +66,7 @@ class CronJob:
     heartbeat: int | None = None
     retries: int | None = None
     ttl: int | None = None
+    kwargs: dict[str, t.Any] | None = None
 
 
 @dataclasses.dataclass
@@ -57,26 +74,33 @@ class Job:
     """
     Main job class representing a run of a function.
 
-    User Provided Arguments
-        function: the async function name to run
-        kwargs: kwargs to pass to the function
-        queue: the saq.Queue object associated with the job
-        key: unique identifier of a job, defaults to uuid1, can be passed in to avoid duplicate jobs
-        timeout: the maximum amount of time a job can run for in seconds, defaults to 10 (0 means disabled)
-        heartbeat: the maximum amount of time a job can survive without a heartbeat in seconds, defaults to 0 (disabled)
+    **User specified Arguments:**
+
+    Args:
+        function (str): the async function name to run
+        kwargs (dict[str, Any] | None): kwargs to pass to the function
+        queue (saq.queue.Queue): the saq.Queue object associated with the job
+        key (str): unique identifier of a job, defaults to uuid1, can be passed in to avoid duplicate jobs
+        timeout (int): the maximum amount of time a job can run for in seconds, defaults to 10 (0 means disabled)
+        heartbeat (int): the maximum amount of time a job can survive without a heartbeat in seconds, defaults to 0 (disabled)
             a heartbeat can be triggered manually within a job by calling await job.update()
-        retries: the maximum number of attempts to retry a job, defaults to 1
-        ttl: the maximum time in seconds to store information about a job including results, defaults to 600 (0 means indefinitely, -1 means disabled)
-        retry_delay: seconds to delay before retrying the job
-        retry_backoff: If true, use exponential backoff for retry delays.
+        retries (int): the maximum number of attempts to retry a job, defaults to 1
+        ttl (int): the maximum time in seconds to store information about a job including results, defaults to 600
+            (0 means indefinitely, -1 means disabled)
+        retry_delay (float): seconds to delay before retrying the job
+        retry_backoff (bool | float): If true, use exponential backoff for retry delays.
             The first retry will have whatever retry_delay is.
             The second retry will have retry_delay*2. The third retry will have retry_delay*4. And so on.
             This always includes jitter, where the final retry delay is a random number between 0 and the calculated retry delay.
             If retry_backoff is set to a number, that number is the maximum retry delay, in seconds.
-        scheduled: epoch seconds for when the job should be scheduled, defaults to 0 (schedule right away)
-        progress: job progress 0.0..1.0
-        meta: arbitrary metadata to attach to the job
-    Framework Set Properties
+        scheduled (int): epoch seconds for when the job should be scheduled, defaults to 0 (schedule right away)
+        progress (float): job progress 0.0..1.0
+        meta (dict): arbitrary metadata to attach to the job
+
+    **Framework Set Properties:**
+    Don't set these, but you can read them.
+
+    Parameters:
         attempts: number of attempts a job has had
         completed: job completion time epoch seconds
         queued: job enqueued time epoch seconds
@@ -85,6 +109,8 @@ class Job:
         result: payload containing the results, this is the return of the function provided, must be serializable, defaults to json
         error: stack trace if a runtime error occurs
         status: Status Enum, default to Status.New
+        priority: The priority of a job, only available in postgres.
+        group_key: Only one job per group can be active at any time, only available in postgres.
     """
 
     function: str
@@ -107,9 +133,30 @@ class Job:
     result: t.Any = None
     error: str | None = None
     status: Status = Status.NEW
+    priority: int = 0
+    group_key: str | None = None
     meta: dict[t.Any, t.Any] = dataclasses.field(default_factory=dict)
 
-    def __repr__(self) -> str:
+    _EXCLUDE_NON_FULL = {
+        "kwargs",
+        "scheduled",
+        "progress",
+        "total_ms",
+        "result",
+        "error",
+        "status",
+        "meta",
+    }
+
+    def info(self, full: bool = False) -> str:
+        """
+        String with Job info
+
+        Args:
+            full: If true, will list the full kwargs for the Job, else an abridged version.
+        """
+        # Using an exclusion list preserves order for kwargs below
+        excluded = set() if full else self._EXCLUDE_NON_FULL
         kwargs = ", ".join(
             f"{k}={v}"
             for k, v in {
@@ -128,26 +175,32 @@ class Job:
                 "status": self.status,
                 "meta": self.meta,
             }.items()
-            if v is not None
+            if v is not None and k not in excluded
         )
         return f"Job<{kwargs}>"
 
+    def __repr__(self) -> str:
+        return self.info(True)
+
     def __hash__(self) -> int:
         return hash(self.key)
 
+    def __eq__(self, other: t.Any) -> bool:
+        return isinstance(other, Job) and self.key == other.key
+
     @property
     def id(self) -> str:
+        """Full Job ID"""
         return self.get_queue().job_id(self.key)
 
-    @classmethod
-    def key_from_id(cls, job_id: str) -> str:
-        return job_id.split(":")[-1]
-
     @property
     def abort_id(self) -> str:
         return f"{ABORT_ID_PREFIX}{self.key}"
 
     def to_dict(self) -> dict[str, t.Any]:
+        """
+        Serialises the Job to dict
+        """
         result = {}
         for field in dataclasses.fields(self):
             key = field.name
@@ -165,8 +218,12 @@ class Job:
         """
         Returns the duration of the job given kind.
 
-        Kind can be process (how long it took to process),
-        start (how long it took to start), or total.
+        Args:
+            Kind (DurationKind): The kind of duration type, can be:
+                * `process` (how long it took to process)
+                * `start` (how long it took to start)
+                * `total`
+                * `running`
         """
         if kind == "process":
             return self._duration(self.completed, self.started)
@@ -185,11 +242,15 @@ class Job:
     def stuck(self) -> bool:
         """Checks if an active job is passed its timeout or heartbeat."""
         current = now()
-        return (self.status == Status.ACTIVE) and bool(
+        return (self.status == Status.ACTIVE or self.status == Status.ABORTING) and bool(
             (self.timeout and seconds(current - self.started) > self.timeout)
             or (self.heartbeat and seconds(current - self.touched) > self.heartbeat)
         )
 
+    @property
+    def retryable(self) -> bool:
+        return self.retries > self.attempts
+
     def next_retry_delay(self) -> float:
         if self.retry_backoff is not False:
             max_delay = None if self.retry_backoff is True else self.retry_backoff
@@ -231,17 +292,18 @@ class Job:
         Updates the stored job in redis.
 
         Set properties with passed in kwargs.
+
+        If status is not explicitly passed in, the status will not update as this is usually controlled by the workers.
         """
-        for k, v in kwargs.items():
-            setattr(self, k, v)
-        await self.get_queue().update(self)
+        await self.get_queue().update(self, **kwargs)
 
     async def refresh(self, until_complete: float | None = None) -> None:
         """
         Refresh the current job with the latest data from the db.
 
-        until_complete: None or Numeric seconds. if None (default), don't wait,
-            else wait seconds until the job is complete or the interval has been reached. 0 means wait forever
+        Args:
+            until_complete (float | None): None or Numeric seconds. if None (default), don't wait,
+                else wait seconds until the job is complete or the interval has been reached. 0 means wait forever
         """
         job = await self.get_queue().job(self.key)
 
@@ -252,7 +314,7 @@ class Job:
 
         if until_complete is not None and not self.completed:
 
-            async def callback(_id: str, status: Status) -> bool:
+            def callback(_id: str, status: Status) -> bool:
                 return status in TERMINAL_STATUSES
 
             await self.get_queue().listen([self.key], callback, until_complete)
